from enum import IntEnum
from typing import List

from pydantic import BaseModel, constr


class MLModelTag(IntEnum):
    archived = 2
    production = 1
    staging = 0
    none = -1


class MLModelIn(BaseModel):
    name: constr(to_lower=True)
    run_id: int


class MLModel(BaseModel):
    run_id: int
    version: str
    name: str
    tag: int
    creation_time: float
    last_modification: float

    class Config:
        orm_mode = True


class MLModelRegistry(BaseModel):
    name: constr(to_lower=True)
    ml_models: List[MLModel]

    class Config:
        orm_mode = True



from typing import List, Optional

from fastapi import APIRouter, HTTPException, UploadFile, File, Query

from huoguoml.schema.run import RunIn, Run
from huoguoml.server.db.service import Service


class RunRouter(object):
    def __init__(self, service: Service):
        router = APIRouter(
            prefix="/api/runs",
            tags=["runs"],
        )

        @router.get("", response_model=List[Run])
        async def get_runs(experiment_name: Optional[str] = Query(None),
                           run_nrs: Optional[str] = Query(None,
                                                          description="String that contains a set of comma seperated "
                                                                      "numbers e.g. '1,2,3,4'")):
            if run_nrs is not None:
                numbers = []
                for el in run_nrs.split(','):
                    el = el.strip()
                    if not el.isnumeric():
                        raise HTTPException(status_code=422,
                                            detail="Element {} in {} is not a number".format(el, run_nrs))
                    numbers.append(int(el))
                run_nrs = numbers
            runs = service.get_runs(experiment_name=experiment_name,
                                    run_nrs=run_nrs)
            return runs

        @router.get("/{run_id}", response_model=Run)
        async def get_run(run_id: int):
            run = service.get_run(run_id=run_id)
            if run is None:
                raise HTTPException(status_code=404)
            return run

        @router.post("", response_model=Run)
        async def create_run(run_in: RunIn):
            return service.create_run(run_in=run_in)

        @router.put("/{run_id}", response_model=Run)
        async def update_run(run_id: int, run: Run):
            run = service.update_run(run_id=run_id, run=run)
            if run is None:
                raise HTTPException(status_code=404)
            return run

        @router.put("/{run_id}/files")
        async def update_or_create_run_files(run_id: int, files: List[UploadFile] = File(...)):
            return service.update_or_create_run_files(run_id=run_id, files=files)

        self.router = router



"""
The huoguoml package provides a easy way for managing Machine Learning types and
services
"""

# DO NOT UPDATE THIS MANUALLY
__version__ = "1624103893"
__license__ = "Apache-2.0"

from huoguoml.tracking import start_experiment_run


from pydantic import BaseModel


class MLServiceIn(BaseModel):
    host: str
    port: int
    model_name: str
    model_rule: str


class MLService(MLServiceIn):
    id: int
    model_version: str

    class Config:
        orm_mode = True


class RequestLog(BaseModel):
    timestamp: int
    sender_host: str
    sender_port: int
    receiver_host: str
    receiver_port: int
    level: int
    message: str

    class Config:
        orm_mode = True



"""
The huoguoml.models module contains the model definition for our ORM mapper SQL Alchemy
"""

from sqlalchemy import Column, Integer, String, ForeignKey, Float, PickleType, desc
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, validates

Base = declarative_base()


class RunORM(Base):
    __tablename__ = "runs"

    id = Column(Integer, primary_key=True, index=True, unique=True)
    run_nr = Column(Integer, index=True)

    creation_time = Column(Float)
    finish_time = Column(Float)
    last_modification = Column(Float)
    duration = Column(Float)

    status = Column(Integer)
    description = Column(String)
    author = Column(String)

    parameters = Column(PickleType)
    metrics = Column(PickleType)
    tags = Column(PickleType)
    model_definition = Column(PickleType)

    experiment = relationship("ExperimentORM", back_populates="runs")
    experiment_name = Column(String, ForeignKey("experiments.name"))

    ml_model = relationship("MLModelORM", uselist=False, back_populates="run")

    @validates('experiment_name')
    def convert_upper(self, key, value):
        self.lower = value.lower()
        return self.lower


class ExperimentORM(Base):
    __tablename__ = "experiments"

    id = Column(Integer, primary_key=True, index=True)
    description = Column(String)

    name = Column(String, index=True, unique=True, nullable=False)
    runs = relationship("RunORM", back_populates="experiment", order_by="desc(RunORM.run_nr)")

    @validates('name')
    def convert_upper(self, key, value):
        self.lower = value.lower()
        return self.lower


class MLServiceORM(Base):
    __tablename__ = "ml_services"

    id = Column(Integer, primary_key=True, index=True)
    host = Column(String)
    port = Column(Integer)
    model_rule = Column(String)
    model_version = Column(String)
    model_name = Column(String)

    model_id = Column(String, ForeignKey('ml_models.id'), nullable=False)
    model = relationship("MLModelORM", back_populates="ml_services")


class MLModelORM(Base):
    __tablename__ = "ml_models"

    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, index=True)
    version = Column(Integer, index=True)
    tag = Column(Integer)
    creation_time = Column(Float)
    last_modification = Column(Float)

    run_id = Column(Integer, ForeignKey('runs.id'), unique=True)
    run = relationship("RunORM", back_populates="ml_model")

    ml_services = relationship("MLServiceORM",
                               back_populates="model")

    @validates('name')
    def convert_upper(self, key, value):
        self.lower = value.lower()
        return self.lower



"""
The huoguoml.util.string module contains some utility functions for string operations.
"""


def create_hash(value: str, algorithm: str) -> str:
    """
    Returns the hash value for a given string value with the given algorithm.
    Supported hash algorithms can be seen with following lines of code:

        import hashlib
        print(hashlib.algorithms_available)
    """
    encoded_value = value.encode("utf-8")
    hash_object = hashlib.new(name=algorithm, data=encoded_value)
    hash_value = hash_object.hexdigest()
    return hash_value


def concat_uri(*args: str):
    return "/".join(args)


def coerce_url(url: str) -> str:
    """
    Gets a url as string and returns it in the right format

    Args:
        url: str

    Returns:
        formatted url as string
    """
    url = url.strip()
    if url.endswith("/"):
        url = url[:-1]

    for proto in ["http://", "https://"]:
        if url.startswith(proto):
            return url
    return "http://{0}".format(url)



"""
The huoguoml.constants module contains used throughout the whole package
"""
from typing import List

import huoguoml

HUOGUOML_METADATA_FILE: str = "huoguoml.meta"
HUOGUOML_DATABASE_FILE: str = "huoguoml.db"

HUOGUOML_DEFAULT_REQUIREMENTS: List[str] = ["huoguoml={}".format(huoguoml.__version__)]


HUOGUOML_DEFAULT_MODEL_FOLDER: str = "model"
HUOGUOML_DEFAULT_ZIP_FOLDER: str = "zips"
HUOGUOML_DEFAULT_FOLDER: str = ".huoguoml"

# SERVICE
HUOGUOML_DEFAULT_SERVICE_PORT: int = 5000
HUOGUOML_DEFAULT_SERVICE_HOST: str = "127.0.0.1"

# SERVER
HUOGUOML_DEFAULT_SERVER_PORT: int = 8080



"""
The huoguoml.util.yaml module contains some utility functions for yaml operations.
"""
from typing import Any

import yaml


def read_yaml(yaml_path: str) -> Any:
    """Read a yaml file and return the content. Returns None if not exist
    """
    with open(yaml_path) as yaml_file:
        data = yaml.load(yaml_file, Loader=yaml.FullLoader)
        return data


def save_yaml(yaml_path: str, data: Any):
    """Save data as yaml format
    """
    with open(yaml_path, 'w') as yaml_file:
        yaml.dump(data, yaml_file)



from typing import List

from fastapi import APIRouter, HTTPException, Request

from huoguoml.schema.ml_service import MLService, MLServiceIn
from huoguoml.server.db.service import Service


class MLServiceRouter(object):
    def __init__(self, service: Service):
        router = APIRouter(
            prefix="/api/services",
            tags=["services"],
        )

        @router.put("", response_model=MLService)
        async def create_or_update_ml_service(ml_service_in: MLServiceIn, request: Request):
            ml_service_in.host = request.client.host
            ml_service = service.create_or_update_ml_service(ml_service_in=ml_service_in)
            if not ml_service:
                raise HTTPException(status_code=400)
            return ml_service

        @router.get("", response_model=List[MLService])
        async def get_ml_services():
            return service.get_ml_services()

        @router.get("/{service_id}", response_model=MLService)
        async def get_ml_service(service_id: int):
            ml_service = service.get_ml_service(service_id=service_id)
            if not ml_service:
                raise HTTPException(status_code=404)
            return ml_service

        self.router = router



"""
The huoguoml.util.string module contains some utility functions for json operations.
"""

import os
import zipfile
from io import BytesIO
from urllib.request import urlopen
from zipfile import ZipFile


def create_zip_file(src_dir: str, dst_dir: str, zip_name: str) -> str:
    """
    Creates a zip file out of a whole directory and saves it under a given name at a given directory
    Args:
        src_dir: source directory
        dst_dir: destination directory
        zip_name: name of the output zip

    Returns:
        zip_file_path: path to the zip file
    """
    zip_file_path = os.path.join(dst_dir, "{}.zip".format(zip_name))
    if not os.path.isfile(zip_file_path):
        zip_file = zipfile.ZipFile(zip_file_path, "w", zipfile.ZIP_DEFLATED)
        abs_src_dir = os.path.abspath(src_dir)
        for dirname, subdirs, files in os.walk(src_dir):
            for filename in files:
                absname = os.path.abspath(os.path.join(dirname, filename))
                arcname = absname[len(abs_src_dir) + 1:]
                zip_file.write(absname, arcname)
        zip_file.close()
    return zip_file_path


def download_and_extract_run_files(run_uri: str, dst_dir: str):
    """
    Gets a URI to a ZIP file, downloads it and extract it to a specific folder.
    """
    with urlopen(run_uri) as zip_file_response:
        with ZipFile(BytesIO(zip_file_response.read())) as zip_file:
            zip_file.extractall(dst_dir)



from typing import List

from pydantic import BaseModel, constr

from huoguoml.schema.run import Run


class ExperimentIn(BaseModel):
    name: constr(to_lower=True)


class Experiment(ExperimentIn):
    """Type for a experiment
    """
    id: int
    description: str
    runs: List[Run]

    class Config:
        orm_mode = True


"""
The huoguoml.cli module contains code for the HuoguoML CLI
"""
import os

import click

from huoguoml.constants import HUOGUOML_DEFAULT_SERVICE_HOST, HUOGUOML_DEFAULT_SERVICE_PORT, HUOGUOML_DEFAULT_FOLDER
from huoguoml.server import start_huoguoml_server
from huoguoml.serving import start_huoguoml_service


@click.group()
@click.version_option()
def cli():
    """
    The HuoguoML CLI. Start the HuoguoML tracking server with 'huoguoml server' or
    a serving with 'huoguoml serving'
    """
    pass


@cli.command()
@click.option(
    "--host",
    default="127.0.0.1",
    help="The network address to listen on (default: 127.0.0.1).",
)
@click.option(
    "--port",
    default=8080,
    help="The port to listen on (default: 8080).",
)
@click.option(
    "--artifact_dir",
    default="/",
    help="The location of the artifact directory for the HuoguoML server (default: /)",
)
def server(host: str, port: int, artifact_dir: str):
    """
    Run the HuoguoML tracking server.
    The server listens on http://127.0.0.1:5000 by default, and only
    accept connections from the local machine. To let the server accept
    connections from other machines, you will need to pass `--host 0.0.0.0`
    to listen on all network interfaces (or a specific interface address).
    """
    start_huoguoml_server(artifact_dir=os.path.join(artifact_dir, HUOGUOML_DEFAULT_FOLDER),
                          host=host,
                          port=port)


@cli.command()
@click.option(
    "--host",
    default=HUOGUOML_DEFAULT_SERVICE_HOST,
    help="The network address to listen on (default: {}).".format(HUOGUOML_DEFAULT_SERVICE_HOST),
)
@click.option(
    "--port",
    default=HUOGUOML_DEFAULT_SERVICE_PORT,
    help="The port to listen on (default: {}).".format(HUOGUOML_DEFAULT_SERVICE_PORT),
)
@click.option(
    "--server_uri",
    help="The uri to the HuoguoML server",
)
@click.option(
    "--model_name",
    help="The name of the model that should be used",
)
@click.option(
    "--model_rule",
    help="The rule for pulling the model e.g. production for always using the production model (default: latest). "
         "Following rules are available: latest, staging, production",
)
@click.option(
    "--artifact_dir",
    default="/",
    help="The location of the artifact directory for the HuoguoML serving. (default: /)",
)
def serving(host: str, port: int, server_uri: str, model_name: str, model_rule: str, artifact_dir: str):
    """
    Run a HuoguoML Serving client.
    The client listens on http://127.0.0.1:8080 by default, and only
    accept connections from the local machine. To let the client accept
    request from other machines, you will need to pass `--host 0.0.0.0`
    to listen on all network interfaces (or a specific interface address).
    """
    start_huoguoml_service(host=host, port=port, server_uri=server_uri, model_name=model_name,
                           model_rule=model_rule, artifact_dir=os.path.join(artifact_dir, HUOGUOML_DEFAULT_FOLDER))



"""
The huoguoml.database module provides the database that contains all informations
"""
import os
from itertools import groupby
from operator import attrgetter
from typing import List, Optional

from fastapi import UploadFile

from huoguoml.constants import HUOGUOML_DATABASE_FILE, HUOGUOML_DEFAULT_ZIP_FOLDER, HUOGUOML_DEFAULT_MODEL_FOLDER, \
    HUOGUOML_METADATA_FILE
from huoguoml.schema.experiment import ExperimentIn, Experiment
from huoguoml.schema.ml_model import MLModelIn, MLModelRegistry, MLModel
from huoguoml.schema.ml_service import MLService, MLServiceIn
from huoguoml.schema.run import Run, RunIn
from huoguoml.server.db.entity import ExperimentORM, RunORM, MLModelORM, MLServiceORM
from huoguoml.server.db.repository import Repository
from huoguoml.util.yaml import save_yaml
from huoguoml.util.zip import create_zip_file


class Service(object):
    """The Repository object is responsible for the connection to the database.
    """

    def __init__(self, artifact_dir: str):
        self.artifact_dir = os.path.realpath(artifact_dir)
        self.zip_dir = os.path.join(self.artifact_dir, HUOGUOML_DEFAULT_ZIP_FOLDER)

        if not os.path.isdir(self.artifact_dir):
            os.makedirs(self.artifact_dir)
            os.makedirs(self.zip_dir)

        database_url = os.path.join("sqlite:///{}".format(artifact_dir), HUOGUOML_DATABASE_FILE)
        connect_args = {"check_same_thread": False}
        self.repository = Repository(database_url=database_url, connect_args=connect_args)

    def get_experiments(self) -> List[ExperimentORM]:
        return self.repository.get_experiments()

    def get_experiment(self, experiment_name: str) -> Optional[ExperimentORM]:
        return self.repository.get_experiment(experiment_name=experiment_name)

    def get_experiment_run(self, experiment_name: str, experiment_run_nr: int) -> Optional[RunORM]:
        return self.repository.get_experiment_run(experiment_name=experiment_name,
                                                  experiment_run_nr=experiment_run_nr)

    def create_experiment(self, experiment_in: ExperimentIn) -> Optional[ExperimentORM]:
        experiment_orm = self.repository.create_experiment(experiment_in=experiment_in)
        os.makedirs(os.path.join(self.artifact_dir, experiment_orm.name))
        return experiment_orm

    def create_run(self, run_in: RunIn) -> RunORM:
        run_orm = self.repository.create_run(run_in=run_in)
        os.makedirs(os.path.join(self.artifact_dir, run_orm.experiment_name, str(run_orm.run_nr)))
        return run_orm

    def get_run(self, run_id: int) -> Optional[RunORM]:
        return self.repository.get_run(run_id=run_id)

    def update_experiment(self, experiment_name: str, experiment: Experiment) -> Optional[ExperimentORM]:
        experiment_orm = self.repository.update_experiment(experiment_name=experiment_name, experiment=experiment)
        return experiment_orm

    def create_or_update_ml_service(self, ml_service_in: MLServiceIn) -> MLService:
        return self.repository.create_or_update_ml_service(ml_service_in=ml_service_in)

    def get_ml_services(self) -> List[MLServiceORM]:
        return self.repository.get_ml_services()

    def update_run(self, run_id: int, run: Run) -> Optional[RunORM]:
        run = self.repository.update_run(run_id=run_id,
                                         run=run)
        run_dir = os.path.join(self.artifact_dir, run.experiment_name, str(run.run_nr))
        save_yaml(yaml_path=os.path.join(run_dir, HUOGUOML_METADATA_FILE),
                  data=Run.from_orm(run).dict())
        return run

    def update_or_create_run_files(self, run_id: int, files: List[UploadFile]) -> bool:
        run_orm = self.repository.get_run(run_id=run_id)
        if run_orm is None:
            return False

        for file in files:
            file_path = os.path.join(self.artifact_dir,
                                     run_orm.experiment_name,
                                     str(run_orm.run_nr),
                                     HUOGUOML_DEFAULT_MODEL_FOLDER,
                                     file.filename)
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(file_path, "wb+") as file_object:
                file_object.write(file.file.read())
        return True

    def get_ml_model_by_name(self, ml_model_name: str) -> List[MLModelORM]:
        return self.repository.get_ml_model_by_name(ml_model_name=ml_model_name)

    def get_ml_models_groupby_name(self) -> List[MLModelRegistry]:
        ml_models_orm = self.repository.get_ml_models()

        registry_list = []
        for k, g in groupby(ml_models_orm, attrgetter('name')):
            ml_models = list(g)
            registry = MLModelRegistry(
                name=k,
                ml_models=ml_models
            )
            registry_list.append(
                registry
            )
        return registry_list

    def get_runs(self, experiment_name: Optional[str], run_nrs: Optional[List[int]]) -> List[RunORM]:
        if experiment_name is not None and run_nrs is not None:
            return self.repository.get_runs_by_experiment_name_and_run_nrs(experiment_name=experiment_name,
                                                                           run_nrs=run_nrs)
        else:
            return self.repository.get_runs()

    def create_ml_model(self, ml_model_in: MLModelIn) -> Optional[MLModelORM]:
        return self.repository.create_ml_model(ml_model_in=ml_model_in)

    def get_ml_model_files_by_name_and_version(self, ml_model_name: str, ml_model_version: str) -> Optional[str]:
        ml_model = self.repository.get_ml_model_files_by_name_and_version(ml_model_name=ml_model_name,
                                                                          ml_model_version=ml_model_version)
        if ml_model:
            run = ml_model.run
            run_dir = os.path.join(self.artifact_dir, run.experiment_name, str(run.run_nr))
            zip_file_path = create_zip_file(src_dir=run_dir, dst_dir=self.zip_dir, zip_name=str(run.id))
            return zip_file_path

    def get_ml_service(self, service_id: int) -> Optional[MLServiceORM]:
        return self.repository.get_ml_service(service_id=service_id)

    def update_ml_model(self, ml_model_name: str, ml_model_version: str, ml_model: MLModel):
        return self.repository.update_ml_model(ml_model_name=ml_model_name,
                                               ml_model_version=ml_model_version,
                                               ml_model=ml_model)

    def update_ml_services_by_model_name(self, ml_model_name: str, ml_model_rule: str) -> List[MLServiceORM]:
        if ml_model_rule == "latest":
            return self.repository.update_ml_services_by_latest(ml_model_name=ml_model_name)
        elif ml_model_rule == "production":
            return self.repository.update_ml_services_by_production(ml_model_name=ml_model_name)
        elif ml_model_rule == "staging":
            return self.repository.update_ml_services_by_staging(ml_model_name=ml_model_name)
        else:
            return []

    def get_ml_model(self, ml_model_name: str, ml_model_version: str) -> Optional[MLModelORM]:
        return self.repository.get_ml_model(ml_model_name=ml_model_name,
                                            ml_model_version=ml_model_version)



from enum import IntEnum
from typing import Optional, List, Dict, Any, Union

from pydantic import BaseModel, constr

from huoguoml.schema.ml_model import MLModel


class ModelNode(BaseModel):
    """Model Node
    """
    dtype: str
    shape: List[Union[int, None]]


class ModelGraph(BaseModel):
    """Type for a model. Every model consists of a input/output
    node with a certain shape and dtype.
    """
    inputs: Dict[str, ModelNode]
    outputs: Dict[str, ModelNode]


class ModelAPI(BaseModel):
    """Type for the model tracking API.
    """
    module: str
    name: str
    arguments: Dict[str, Any]


class ModelDefinition(BaseModel):
    """Type for the definition of a model"""
    model_graph: ModelGraph
    model_api: ModelAPI
    requirements: List[str]


class RunStatus(IntEnum):
    completed = 1
    failed = 0
    pending = -1


class RunIn(BaseModel):
    experiment_name: constr(to_lower=True)
    author: str
    status: int = RunStatus.pending.value


class Run(RunIn):
    """Type for a single experiment Run
    """
    id: int
    run_nr: int

    creation_time: float
    last_modification: float
    finish_time: float
    duration: float

    description: str

    parameters: Dict[str, str]
    metrics: Dict[str, str]
    tags: Dict[str, str]
    model_definition: Optional[ModelDefinition]

    ml_model: Optional[MLModel]

    class Config:
        orm_mode = True



from typing import List

from fastapi import APIRouter, HTTPException

from huoguoml.schema.experiment import Experiment, ExperimentIn
from huoguoml.schema.run import Run
from huoguoml.server.db.service import Service


class ExperimentRouter(object):

    def __init__(self, service: Service):
        router = APIRouter(
            prefix="/api/experiments",
            tags=["experiments"],
        )

        @router.get("", response_model=List[Experiment])
        async def get_experiments():
            return service.get_experiments()

        @router.post("", response_model=Experiment)
        async def create_experiment(experiment_in: ExperimentIn):
            experiment = service.create_experiment(experiment_in=experiment_in)
            if experiment is None:
                raise HTTPException(status_code=422)
            return experiment

        @router.get("/{experiment_name}", response_model=Experiment)
        async def get_experiment(experiment_name: str):
            experiment = service.get_experiment(experiment_name=experiment_name)
            if experiment is None:
                raise HTTPException(status_code=404)
            return experiment

        @router.get("/{experiment_name}/{experiment_run_nr}", response_model=Run)
        async def get_experiment_run(experiment_name: str, experiment_run_nr: int):
            return service.get_experiment_run(experiment_name=experiment_name,
                                              experiment_run_nr=experiment_run_nr)

        @router.put("/{experiment_name}", response_model=Experiment)
        async def update_experiment(experiment_name: str, experiment: Experiment):
            experiment = service.update_experiment(experiment_name=experiment_name, experiment=experiment)
            if experiment is None:
                raise HTTPException(status_code=404)
            return experiment

        self.router = router



"""
The huoguoml.database module provides the database that contains all informations
"""
import time
from typing import List, Dict, Optional

from sqlalchemy import create_engine, desc
from sqlalchemy.orm import sessionmaker, scoped_session

from huoguoml.schema.experiment import ExperimentIn, Experiment
from huoguoml.schema.ml_model import MLModelIn, MLModelTag, MLModel
from huoguoml.schema.ml_service import MLServiceIn
from huoguoml.schema.run import RunIn, Run
from huoguoml.server.db.entity import Base, RunORM, MLServiceORM, ExperimentORM, MLModelORM


class Repository(object):
    """The Repository object is responsible for the connection to the database.
    """

    def __init__(self, database_url: str, connect_args: Dict):
        engine = create_engine(
            database_url, connect_args=connect_args
        )
        session_factory = sessionmaker(bind=engine)
        self.Session = scoped_session(session_factory)
        Base.metadata.create_all(bind=engine)

    # Experiment
    def get_experiments(self) -> List[ExperimentORM]:
        session = self.Session()
        return session.query(ExperimentORM).all()

    def get_experiment(self, experiment_name: str) -> Optional[ExperimentORM]:
        session = self.Session()
        experiment = session.query(ExperimentORM).filter_by(name=experiment_name).first()
        return experiment

    def create_experiment(self, experiment_in: ExperimentIn) -> Optional[ExperimentORM]:
        session = self.Session()
        experiment = session.query(ExperimentORM).filter_by(name=experiment_in.name).first()
        if experiment:
            return None

        experiment = ExperimentORM(
            description="",
            **experiment_in.dict())
        session.add(experiment)
        session.commit()
        session.refresh(experiment)
        return experiment

    # Run
    def create_run(self, run_in: RunIn) -> RunORM:
        session = self.Session()
        experiment = session.query(ExperimentORM).filter_by(name=run_in.experiment_name).first()

        run = RunORM(
            run_nr=len(experiment.runs) + 1,
            creation_time=time.time(),
            last_modification=time.time(),
            finish_time=-1,
            duration=-1,
            description="",
            parameters={},
            metrics={},
            tags={},
            model_definition=None,
            **run_in.dict()
        )

        session.add(run)
        session.commit()
        session.refresh(run)
        return run

    def get_run(self, run_id: int) -> Optional[RunORM]:
        session = self.Session()
        return session.query(RunORM).filter_by(id=run_id).first()

    def update_experiment(self, experiment_name: str, experiment: Experiment) -> Optional[ExperimentORM]:
        session = self.Session()
        experiment_orm = session.query(ExperimentORM).filter_by(name=experiment_name, id=experiment.id).first()
        if experiment_orm:
            for field, field_value in experiment.dict(exclude_unset=True, exclude={"id", "runs"}).items():
                setattr(experiment_orm, field, field_value)

            session.commit()
            session.refresh(experiment_orm)
            return experiment_orm

    def update_run(self, run_id: int, run: Run) -> Optional[RunORM]:
        session = self.Session()
        run_orm = session.query(RunORM).filter_by(id=run_id).first()

        if run_orm:
            for field, field_value in run.dict(exclude={"id", "run_nr", "experiment_name", "ml_model"}).items():
                setattr(run_orm, field, field_value)
            run_orm.last_modification = time.time()
            session.commit()
            session.refresh(run_orm)
            return run_orm

    def get_experiment_run(self, experiment_name: str, experiment_run_nr: int) -> Optional[RunORM]:
        session = self.Session()
        return session.query(RunORM).filter_by(run_nr=experiment_run_nr, experiment_name=experiment_name
                                               ).first()

    def get_runs_by_experiment_name_and_run_nrs(self, experiment_name: str, run_nrs: List[int]) -> List[RunORM]:
        session = self.Session()
        return session.query(RunORM).filter(experiment_name == RunORM.experiment_name,
                                            RunORM.run_nr.in_(run_nrs)).all()

    def get_runs(self) -> List[RunORM]:
        session = self.Session()
        return session.query(RunORM).all()

    # MLModel
    def get_ml_models(self):
        session = self.Session()
        return session.query(MLModelORM).order_by(MLModelORM.name).all()

    def get_ml_model_by_name(self, ml_model_name) -> List[MLModelORM]:
        session = self.Session()
        return session.query(MLModelORM).filter_by(name=ml_model_name).order_by(desc(MLModelORM.version)).all()

    def create_ml_model(self, ml_model_in: MLModelIn) -> Optional[MLModelORM]:
        session = self.Session()
        run_orm = session.query(RunORM).filter_by(id=ml_model_in.run_id).first()
        if not run_orm or run_orm and run_orm.ml_model:
            return None

        version = session.query(MLModelORM).filter_by(name=ml_model_in.name).count() + 1
        ml_model = MLModelORM(
            tag=MLModelTag.none.value,
            version="v{}".format(version),
            creation_time=time.time(),
            last_modification=time.time(),
            **ml_model_in.dict())
        session.add(ml_model)
        session.commit()
        session.refresh(ml_model)
        return ml_model

    def get_ml_model_by_name_and_tag(self, ml_model_name: str, tag: int) -> Optional[MLModelORM]:
        session = self.Session()
        return session.query(MLModelORM).filter_by(name=ml_model_name,
                                                   tag=tag).first()

    def get_ml_model_by_name_and_latest(self, ml_model_name: str) -> Optional[MLModelORM]:
        session = self.Session()
        return session.query(MLModelORM).filter_by(name=ml_model_name).order_by(desc(MLModelORM.version)).first()

    def get_ml_model_files_by_name_and_version(self, ml_model_name: str, ml_model_version: str) -> Optional[MLModelORM]:
        session = self.Session()
        return session.query(MLModelORM).filter_by(name=ml_model_name,
                                                   version=ml_model_version).first()

    # Services
    def create_or_update_ml_service(self, ml_service_in: MLServiceIn) -> Optional[MLServiceORM]:
        session = self.Session()

        ml_model = None
        if ml_service_in.model_rule == "production":
            ml_model = self.get_ml_model_by_name_and_tag(ml_model_name=ml_service_in.model_name,
                                                         tag=MLModelTag.staging.value)
        elif ml_service_in.model_rule == "latest":
            ml_model = self.get_ml_model_by_name_and_latest(ml_model_name=ml_service_in.model_name)
        elif ml_service_in.model_rule == "staging":
            ml_model = self.get_ml_model_by_name_and_tag(ml_model_name=ml_service_in.model_name,
                                                         tag=MLModelTag.staging.value)
        if ml_model:
            ml_service = session.query(MLServiceORM).filter_by(host=ml_service_in.host,
                                                               port=ml_service_in.port).first()
            if ml_service:
                ml_service.model_name = ml_model.name
                ml_service.model_version = ml_model.version
                ml_service.model = ml_model
            else:
                ml_service = MLServiceORM(
                    host=ml_service_in.host,
                    port=ml_service_in.port,
                    model=ml_model,
                    model_name=ml_model.name,
                    model_version=ml_model.version,
                    model_rule=ml_service_in.model_rule,
                )
                session.add(ml_service)
            session.commit()
            session.refresh(ml_service)
            return ml_service

    def get_ml_services(self):
        session = self.Session()
        return session.query(MLServiceORM).all()

    def get_ml_service(self, service_id: int) -> Optional[MLServiceORM]:
        session = self.Session()
        return session.query(MLServiceORM).filter_by(id=service_id).first()

    def update_ml_services_by_latest(self, ml_model_name: str) -> List[MLServiceORM]:
        session = self.Session()
        ml_services = session.query(MLServiceORM).filter_by(model_name=ml_model_name,
                                                            model_rule="latest").all()
        ml_model = self.get_ml_model_by_name_and_latest(ml_model_name=ml_model_name)

        for ml_service in ml_services:
            ml_service.model_name = ml_model.name
            ml_service.model_version = ml_model.version
            ml_service.model = ml_model

            session.commit()
            session.refresh(ml_service)
        return ml_services

    def update_ml_services_by_staging(self, ml_model_name: str) -> List[MLServiceORM]:
        session = self.Session()
        ml_services = session.query(MLServiceORM).filter_by(model_name=ml_model_name,
                                                            model_rule="staging").all()
        ml_model = self.get_ml_model_by_name_and_tag(ml_model_name=ml_model_name, tag=MLModelTag.staging.value)
        for ml_service in ml_services:
            ml_service.model_name = ml_model.name
            ml_service.model_version = ml_model.version
            ml_service.model = ml_model
            session.commit()
            session.refresh(ml_service)
        return ml_services

    def update_ml_services_by_production(self, ml_model_name: str) -> List[MLServiceORM]:
        session = self.Session()
        ml_services = session.query(MLServiceORM).filter_by(model_name=ml_model_name,
                                                            model_rule="production").all()
        ml_model = self.get_ml_model_by_name_and_tag(ml_model_name=ml_model_name, tag=MLModelTag.production.value)

        for ml_service in ml_services:
            ml_service.model_name = ml_model.name
            ml_service.model_version = ml_model.version
            ml_service.model = ml_model

            session.commit()
            session.refresh(ml_service)
        return ml_services

    def get_ml_model(self, ml_model_name: str, ml_model_version: str) -> Optional[MLModelORM]:
        session = self.Session()
        return session.query(MLModelORM).filter_by(name=ml_model_name,
                                                   version=ml_model_version).first()

    def update_ml_model(self, ml_model_name: str, ml_model_version: str, ml_model: MLModel) -> Optional[MLModelORM]:
        session = self.Session()

        ml_model_orm = session.query(MLModelORM).filter_by(name=ml_model_name,
                                                           version=ml_model_version).first()
        if ml_model_orm:
            temp = session.query(MLModelORM).filter_by(name=ml_model_name, tag=ml_model.tag).first()
            if temp and (ml_model.tag == MLModelTag.staging.value or ml_model.tag == MLModelTag.production.value):
                temp.tag = MLModelTag.none.value

            for field, field_value in ml_model.dict(exclude={"id", "run_id", "version", "name"}).items():
                setattr(ml_model_orm, field, field_value)
            session.commit()
            session.refresh(ml_model_orm)

        return ml_model_orm



"""
The huoguoml.tracking module provides the options for tracking all experiment runs
"""
import getpass
import time
from typing import List, Optional

import requests

from huoguoml.schema.experiment import Experiment, ExperimentIn
from huoguoml.schema.run import Run, RunIn, RunStatus
from huoguoml.util.string import concat_uri, coerce_url


class HuoguoMLRun(object):

    def __init__(self,
                 experiment_name: str,
                 server_uri: str):
        self.server_uri = server_uri

        experiment_api = concat_uri(self.server_uri, "api", "experiments")
        exp_res = requests.get(concat_uri(experiment_api, experiment_name))
        if exp_res.status_code.real == 404:
            exp_in = ExperimentIn(name=experiment_name)
            exp_res = requests.post(
                experiment_api,
                data=exp_in.json())
        experiment = Experiment.parse_raw(exp_res.text)

        run_in = RunIn(experiment_name=experiment.name, author=getpass.getuser())
        run_res = requests.post(
            concat_uri(self.server_uri, "api", "runs"),
            data=run_in.json())

        self._run = Run.parse_raw(run_res.text)

    def log_model(self,
                  model_type: str,
                  **kwargs):
        if self._run.model_definition:
            raise FileExistsError("Model already logged")

        if model_type == "tensorflow":
            from huoguoml.tracking.tensorflow import log_model
            model_definition, model_files = log_model(**kwargs)
        else:
            raise NotImplementedError("HuoguoML does not support {} at the moment".format(model_type))

        self._run.model_definition = model_definition
        self._update_experiment_run(model_files)

    def log_parameter(self, parameter_name: str, parameter_value: str):
        self._run.parameters[parameter_name] = parameter_value
        self._update_experiment_run()

    def log_metric(self, metric_name: str, metric_value: str):
        self._run.metrics[metric_name] = metric_value
        self._update_experiment_run()

    def log_tag(self, tag_name: str, tag_value: str):
        self._run.tags[tag_name] = tag_value
        self._update_experiment_run()

    def end_experiment_run(self, failed=False):
        self._run.finish_time = time.time()
        self._run.duration = self._run.finish_time - self._run.creation_time

        if failed:
            self._run.status = RunStatus.failed.value
        else:
            self._run.status = RunStatus.completed.value

        self._update_experiment_run()

    def _update_experiment_run(self, files: Optional[List] = None):
        api = concat_uri(self.server_uri, "api", "runs", str(self._run.id))
        if files:
            requests.put(concat_uri(api, "files"), files=files)

        requests.put(
            api,
            data=self._run.json())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_value:
            self.end_experiment_run(failed=True)
        else:
            self.end_experiment_run(failed=False)


def start_experiment_run(experiment_name: str,
                         server_uri: str = "127.0.0.1:8080"):
    """
    Start a HuoguoML experiments run and returns it. The return value can be used as a context manager within a with
    block; otherwise, you must call end_experiment_run() to terminate the current run. In addition the HuoguoML server
    must be running.

    Args:
        experiment_name: Name of the experiment under which to create the experiment run
        server_uri: Uri to the HuoguoML server. (default: 127.0.0.1:8080)
    """
    server_uri = coerce_url(server_uri)

    server_res = requests.get(server_uri)
    if server_res.status_code >= 400:
        raise ConnectionError(
            "HuoguoML server cannot be found on {}. Start server with 'huoguoml server' and or try another server uri".format(
                server_uri))

    return HuoguoMLRun(server_uri=server_uri, experiment_name=experiment_name)



from typing import List

import requests
from fastapi import APIRouter, HTTPException
from starlette.background import BackgroundTasks
from starlette.responses import FileResponse

from huoguoml.schema.ml_model import MLModelIn, MLModel, MLModelRegistry, MLModelTag
from huoguoml.schema.ml_service import MLService
from huoguoml.server.db.service import Service
from huoguoml.util.string import concat_uri, coerce_url


class MLModelRouter(object):

    def __init__(self, service: Service):
        router = APIRouter(
            prefix="/api/models",
            tags=["models"],
        )

        def notify_services(ml_model_name: str, ml_model_rule: str):
            ml_services = service.update_ml_services_by_model_name(ml_model_name=ml_model_name,
                                                                   ml_model_rule=ml_model_rule)
            for ml_service in ml_services:
                uri = "{}:{}".format(coerce_url(ml_service.host), ml_service.port)
                ml_service_obj = MLService.from_orm(ml_service)
                update_api = concat_uri(uri, "api", "update")
                requests.post(update_api, json=ml_service_obj.dict()).raise_for_status()

        @router.get("", response_model=List[MLModelRegistry])
        async def get_ml_models_groupby_name():
            return service.get_ml_models_groupby_name()

        @router.post("", response_model=MLModel)
        async def create_ml_model(ml_model_in: MLModelIn, background_tasks: BackgroundTasks):
            ml_model = service.create_ml_model(ml_model_in=ml_model_in)
            background_tasks.add_task(notify_services, ml_model_name=ml_model_in.name, ml_model_rule="latest")
            if ml_model is None:
                raise HTTPException(status_code=400)
            return ml_model

        @router.put("/{ml_model_name}/{ml_model_version}", response_model=MLModel)
        async def update_ml_model(ml_model_name: str, ml_model_version: str, ml_model: MLModel,
                                  background_tasks: BackgroundTasks):
            ml_model = service.update_ml_model(ml_model_name=ml_model_name,
                                               ml_model_version=ml_model_version,
                                               ml_model=ml_model)

            if ml_model.tag == MLModelTag.production.value:
                background_tasks.add_task(notify_services, ml_model_name=ml_model.name, ml_model_rule="production")
            elif ml_model.tag == MLModelTag.staging.value:
                background_tasks.add_task(notify_services, ml_model_name=ml_model.name, ml_model_rule="staging")
            else:
                raise HTTPException(status_code=400)

            if ml_model is None:
                raise HTTPException(status_code=400)
            return ml_model

        @router.get("/{ml_model_name}", response_model=List[MLModel])
        async def get_ml_models(ml_model_name: str):
            ml_model = service.get_ml_model_by_name(ml_model_name=ml_model_name)
            if not ml_model:
                raise HTTPException(status_code=404)
            return ml_model

        @router.get("/{ml_model_name}/{ml_model_version}", response_model=MLModel)
        async def get_ml_model(ml_model_name: str, ml_model_version: str):
            ml_model = service.get_ml_model(ml_model_name=ml_model_name,
                                            ml_model_version=ml_model_version)
            if not ml_model:
                raise HTTPException(status_code=404)

            return ml_model

        @router.get("/{ml_model_name}/{ml_model_version}/files")
        async def get_ml_model_files(ml_model_name: str, ml_model_version: str):
            file_path = service.get_ml_model_files_by_name_and_version(ml_model_name=ml_model_name,
                                                                       ml_model_version=ml_model_version)
            if not file_path:
                raise HTTPException(status_code=404)

            return FileResponse(file_path, media_type='application/zip')

        self.router = router



import os

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette.staticfiles import StaticFiles

from huoguoml.server.api.experiment import ExperimentRouter
from huoguoml.server.api.ml_model import MLModelRouter
from huoguoml.server.api.ml_service import MLServiceRouter
from huoguoml.server.api.run import RunRouter
from huoguoml.server.db.service import Service


def start_huoguoml_server(artifact_dir: str, host: str, port: int):
    """
    Starts the HuoguoML server

    Args:
        artifact_dir: Location of the artifact directory
        host: The network address to listen on
        port: The port to listen on
    """
    service = Service(artifact_dir=artifact_dir)

    app = FastAPI()

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    app.include_router(ExperimentRouter(service=service).router)
    app.include_router(RunRouter(service=service).router)
    app.include_router(MLServiceRouter(service=service).router)
    app.include_router(MLModelRouter(service=service).router)

    dashboard_files_dir = os.path.join(os.path.dirname(__file__),
                                       "dashboard",
                                       "build")

    app.mount("/",
              StaticFiles(directory=dashboard_files_dir, html=True),
              name="static")

    uvicorn.run(app, host=host, port=port)


if __name__ == '__main__':
    start_huoguoml_server("../../examples/huoguoml", "127.0.0.1", 8080)



"""
The huoguoml.tracking module provides the options for tracking tensorflow experiments
"""
import os
from typing import List, Tuple

from huoguoml.constants import HUOGUOML_DEFAULT_REQUIREMENTS, HUOGUOML_DEFAULT_MODEL_FOLDER
from huoguoml.schema.run import ModelNode, ModelDefinition, ModelAPI, ModelGraph


def get_requirements() -> List[str]:
    """
    Returns a list of requirements that are required for a Tensorflow tracking.
    """
    import tensorflow as tf

    requirements = HUOGUOML_DEFAULT_REQUIREMENTS.copy()
    requirements.append("tensorflow={}".format(tf.__version__))
    return requirements


def get_file_pattern() -> List[str]:
    """
    Returns a list with all possible file patterns
    """
    return ["*.pb", "*.data", "*.index"]


def _load_saved_model(tf_saved_model_dir: str, tf_meta_graph_tags: str, tf_signature_def_key: str):
    """
    Loads a TF SavedModel
    """
    import tensorflow as tf

    model = tf.saved_model.load(
        tags=tf_meta_graph_tags, export_dir=tf_saved_model_dir
    )
    signatures = model.signatures

    if tf_signature_def_key not in signatures:
        raise KeyError(
            "Could not find signature def key %s. Available keys are: %s"
            % (tf_signature_def_key, list(signatures.keys()))
        )
    return signatures[tf_signature_def_key]


def log_model(
        tf_saved_model_dir: str,
        tf_meta_graph_tags: str,
        tf_signature_def_key: str) -> Tuple[ModelDefinition, List]:
    import tensorflow as tf
    if tf.__version__ < "2.0.0":
        raise NotImplementedError("HuoguoML does not support Tensorflow 1.X")

    # check if saved model is valid
    saved_model = _load_saved_model(
        tf_saved_model_dir=tf_saved_model_dir,
        tf_meta_graph_tags=tf_meta_graph_tags,
        tf_signature_def_key=tf_signature_def_key,
    )
    requirements = get_requirements()
    inputs = {}
    for input_tensor in saved_model.structured_input_signature:
        if input_tensor:
            for tensor_name, tensor_spec in input_tensor.items():
                inputs[tensor_name] = ModelNode(shape=tensor_spec.shape.as_list(),
                                                dtype=tensor_spec.dtype.name)
    outputs = {}
    for tensor_name, tensor_spec in saved_model.structured_outputs.items():
        outputs[tensor_name] = ModelNode(shape=tensor_spec.shape.as_list(),
                                         dtype=tensor_spec.dtype.name)
    model_graph = ModelGraph(inputs=inputs, outputs=outputs)

    model_api = ModelAPI(module="tensorflow",
                         arguments={
                             "tf_saved_model_dir": HUOGUOML_DEFAULT_MODEL_FOLDER,
                             "tf_meta_graph_tags": tf_meta_graph_tags,
                             "tf_signature_def_key": tf_signature_def_key,
                         },
                         name="load_model")

    model_files = []
    for dir_path, _, filenames in os.walk(tf_saved_model_dir):
        for filename in filenames:
            file_path = os.path.join(dir_path, filename)
            abs_file_path = os.path.abspath(file_path)
            model_file_path = os.path.relpath(file_path, tf_saved_model_dir)
            model_file = ("files", (model_file_path, open(abs_file_path, "rb")))
            model_files.append(model_file)
    model_definition = ModelDefinition(model_api=model_api,
                                       model_graph=model_graph,
                                       requirements=requirements)
    return model_definition, model_files


def load_model(tf_saved_model_dir: str, tf_meta_graph_tags: str, tf_signature_def_key: str):
    saved_model = _load_saved_model(tf_saved_model_dir=tf_saved_model_dir,
                                    tf_meta_graph_tags=tf_meta_graph_tags,
                                    tf_signature_def_key=tf_signature_def_key)
    return TFModel(saved_model=saved_model)


class TFModel(object):
    """
    TFModel used for predictions
    """

    def __init__(self, saved_model):
        self.saved_model = saved_model

    def predict(self, data):
        import tensorflow as tf

        data_dict = data.dict()
        for key in data_dict.keys():
            data_dict[key] = tf.constant(data_dict[key])

        pred = self.saved_model(**data_dict)
        pred_dict = {col_name: pred[col_name].numpy().tolist() for col_name in pred.keys()}
        return pred_dict



import os

from setuptools import find_packages
from setuptools import setup


def get_extra_requires(path, add_all=True):
    """Returns a dict with all tags as key and a list of their respective packages
    """
    from collections import defaultdict
    import re
    with open(path) as fp:
        extra_deps = defaultdict(set)
        for k in fp:
            if k.strip() and not k.startswith("#"):
                tags = set()
                if ":" in k:
                    k, v = k.split(":")
                    tags.update(vv.strip() for vv in v.split(","))
                tags.add(re.split("[<=>]", k)[0])
                for t in tags:
                    extra_deps[t].add(k)

        # add tag `all` at the end
        if add_all:
            extra_deps["all"] = set(vv for v in extra_deps.values() for vv in v)
    return extra_deps


# get all packages in requirements file
dependencies = get_extra_requires("extra-requirements.txt")

libinfo_py = os.path.join("huoguoml", "__init__.py")
libinfo_content = open(libinfo_py, "r", encoding="utf8").readlines()
version_line = [l.strip() for l in libinfo_content if l.startswith("__version__")][
    0
]
version_dict = {}
exec(version_line, version_dict)

# load the README file and use it as the long_description for PyPI
with open("README.md", encoding="utf8") as f:
    readme = f.read()


def package_files(directory):
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))
    return paths


pkg_name = "huoguoml-dev"

setup(
    name=pkg_name,
    description="A platform for managing and serving Machine Learning types.",
    long_description=readme,
    long_description_content_type="text/markdown",
    version=version_dict["__version__"],
    author="Steven Mi",
    url="https://huoguoml.github.io",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    package_data={"huoguoml": package_files(os.path.join("huoguoml", "server", "dashboard", "build"))},
    python_requires=">=3.7.*",
    install_requires=list(dependencies["all"]),
    extras_require=dependencies,
    license="Apache 2.0",
    zip_safe=False,
    entry_points="""
        [console_scripts]
        huoguoml=huoguoml.cli:cli
    """,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Unix Shell",
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="HuoguoML"
)



import uvicorn
from fastapi import FastAPI
from starlette.middleware.cors import CORSMiddleware

from huoguoml.schema.ml_service import MLServiceIn
from huoguoml.serving.api import HuoguoMLRouter
from huoguoml.util.string import coerce_url


def start_huoguoml_service(host: str, port: int, model_name: str, model_rule: str, server_uri: str, artifact_dir: str):
    """
    Starts the HuoguoML serving

    Args:
        server_uri: The URI to the HuoguoML server
        host: The network address to listen on
        port: The port to listen on
        model_name: The name of the model that should be used
        model_rule: The rule for pulling the model e.g. production for always using the production model
                    (default: latest). Following rules are available: latest, staging, production
        artifact_dir: Location of the artifact directory
    """
    app = FastAPI()
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    ml_service_in = MLServiceIn(host=host, port=port, model_name=model_name, model_rule=model_rule)
    router = HuoguoMLRouter(ml_service_in=ml_service_in, server_uri=coerce_url(server_uri),
                            artifact_dir=artifact_dir).router
    app.include_router(router)

    @app.get("/")
    async def ok():
        return

    uvicorn.run(app, host=host, port=port)


if __name__ == '__main__':
    start_huoguoml_service(host="127.0.0.1", port=5000,
                           server_uri="http://localhost:8080",
                           model_name="mnist",
                           model_rule="latest",
                           artifact_dir="../../examples/huoguoml_service")



"""
The huoguoml.util.string module contains some utility functions for json operations.
"""
import json
from typing import Any


def read_json(json_path: str) -> Any:
    """Read a json file and return the content. Returns None if not exist
    """
    with open(json_path) as json_file:
        data = json.load(json_file)
        return data


def save_json(json_path: str, data: Any):
    """Save data as json format
    """
    with open(json_path, 'w') as json_file:
        json.dump(data, json_file)



import importlib
import os
import shutil
from typing import List

import requests
from fastapi import APIRouter
from pydantic import create_model

from huoguoml.constants import HUOGUOML_METADATA_FILE, HUOGUOML_DEFAULT_FOLDER
from huoguoml.schema.ml_service import MLServiceIn, MLService
from huoguoml.schema.run import Run
from huoguoml.util.string import concat_uri
from huoguoml.util.yaml import read_yaml
from huoguoml.util.zip import download_and_extract_run_files


def load_run_model(run: Run):
    module_name = 'huoguoml.tracking.{}'.format(run.model_definition.model_api.module)
    module = importlib.import_module(module_name)
    function = getattr(module, run.model_definition.model_api.name)
    return function(**run.model_definition.model_api.arguments)


class HuoguoMLRouter(object):
    output_model = None
    input_model = None

    def __init__(self, ml_service_in: MLServiceIn, artifact_dir: str, server_uri: str):
        router = APIRouter(
            prefix="/api",
        )

        register_api = concat_uri(server_uri, "api", "services")
        response = requests.put(register_api, json=ml_service_in.dict())
        self.ml_service = MLService.parse_raw(response.text)
        self.artifact_dir = artifact_dir
        self.server_uri = server_uri

        # TODO: Check if register model is available in artifact_dir, otherwise download
        self._update(ml_service=self.ml_service)

        @router.post("/update", response_model=MLService)
        async def update_model(ml_service: MLService):
            self._update(ml_service)

        @router.post("/predict", response_model=HuoguoMLRouter.output_model)
        async def predict(data: HuoguoMLRouter.input_model):
            return self.model.predict(data)

        @router.get("/version", response_model=MLService)
        async def version():
            return self.ml_service

        self.router = router

    def _update(self, ml_service: MLService):
        self.ml_service = ml_service

        source_dir = os.getcwd()
        model_url = concat_uri(self.server_uri, "api", "models", ml_service.model_name,
                               ml_service.model_version, "files")
        model_dir = os.path.join(HUOGUOML_DEFAULT_FOLDER)

        if os.path.isdir(model_dir):
            shutil.rmtree(model_dir)
        download_and_extract_run_files(run_uri=model_url, dst_dir=model_dir)

        run_dict = read_yaml(os.path.join(model_dir, HUOGUOML_METADATA_FILE))
        run = Run(**run_dict)

        input_model_type = {}
        for input_name, _ in run.model_definition.model_graph.inputs.items():
            input_model_type[input_name] = (List, [])
        HuoguoMLRouter.input_model = create_model('Inputs', **input_model_type)

        output_model_type = {}
        for output_name, _ in run.model_definition.model_graph.outputs.items():
            output_model_type[output_name] = (List, [])
        HuoguoMLRouter.output_model = create_model('Outputs', **output_model_type)

        os.chdir(model_dir)
        self.model = load_run_model(run=run)
        os.chdir(source_dir)



"""
The huoguoml.util.string module contains some utility functions for io operations.
"""
import os


def create_huoguoml_folders(huoguoml_path: str):
    """
    Create HuoguoML folders if not exist

    Args:
        huoguoml_path: path to the huoguoml dir

    NOTE: Required for later stages, when huoguoml folder gets larger
    """
    os.makedirs(huoguoml_path, exist_ok=True)
